==> xkaapi-2.1.tar.gz: xkaapi-2.1/src/perfmodel/kaapi_perfmodel_comm.c <==

#include <stdio.h>
#include <math.h>

#include "kaapi_impl.h"
#include "kaapi_tasklist.h"

#include "kaapi_perfmodel_idgraf.h"

#if defined(KAAPI_USE_CUDA)
#include "machine/cuda/kaapi_cuda_impl.h"
#endif


#define KAAPI_PERFMODEL_COMM_FILE    "kaapi_perfmodel_comm_dump"
#define KAAPI_PERFMODEL_MAX_PARAMS    10

/** return time in microseconds (us)
 */
static inline double kaapi_get_elapsedus(void)
{
  uint64_t retval;
  retval = kaapi_get_elapsedns();
  return (double)(retval / 1e3);
}

typedef struct kaapi_perfmodel_comm_t {
  void      (*init)       (void);
  void      (*finalize)   (void);
  double    (*predict)    (kaapi_processor_t* const, kaapi_taskdescr_t* const);
  void      (*update)     (const kaapi_address_space_id_t, const kaapi_address_space_id_t, const size_t, const double);
} kaapi_perfmodel_comm_t;

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm;

typedef struct kaapi_perfmodel_comm_history_hash_t {
  kaapi_lock_t lock;
  kaapi_big_hashmap_t hash;
}kaapi_perfmodel_comm_history_hash_t;

/* dimensions from source -> destiny */
static kaapi_perfmodel_comm_history_hash_t kaapi_perfmodel_comm_hash[KAAPI_PROC_TYPE_MAX][KAAPI_PROC_TYPE_MAX];

typedef struct kaapi_perfmodel_comm_history_t {
  uint64_t key;
  struct perf {
    kaapi_lock_t lock;
    double mean;    /* mean of times */
    double sd;      /* standard deviation */
    double sum;     /* all sum of execution times */
    double sum_pow2;
    uint64_t samples; /* number of samples */
  } dev[KAAPI_MAX_PROCESSOR_LIMIT][KAAPI_MAX_PROCESSOR_LIMIT];
} kaapi_perfmodel_comm_history_t;

static inline void kaapi_perfmodel_comm_history_init_entry(kaapi_perfmodel_comm_history_t* const perf,
                                                   const uint64_t key)
{
  int i, j;
  perf->key   = key;
  
  for(i= 0; i < KAAPI_MAX_PROCESSOR_LIMIT; i++)
  {
    for(j= 0; j < KAAPI_MAX_PROCESSOR_LIMIT; j++)
    {
      kaapi_atomic_initlock(&perf->dev[i][j].lock);
      perf->dev[i][j].mean     = 0.0f;
      perf->dev[i][j].sd       = 0.0f;
      perf->dev[i][j].sum      = 0.0f;
      perf->dev[i][j].sum_pow2 = 0.0f;
      perf->dev[i][j].samples  =  0;
    }
  }
}

static inline kaapi_perfmodel_comm_history_t* kaapi_perfmodel_comm_history_find_entry(kaapi_perfmodel_comm_history_hash_t* hash, uint64_t key)
{
  kaapi_hashentries_t *entry;
  
  kaapi_atomic_lock(&hash->lock);
  entry = kaapi_big_hashmap_findinsert(&hash->hash, (void*)key);
  kaapi_atomic_unlock(&hash->lock);
  if(entry->u.perf == 0)
  {
    entry->u.perf = (void*)malloc(sizeof(kaapi_perfmodel_comm_history_t));
    kaapi_perfmodel_comm_history_init_entry((kaapi_perfmodel_comm_history_t*)entry->u.perf, key);
  }
  
  return (kaapi_perfmodel_comm_history_t*)entry->u.perf;
}

static inline void kaapi_perfmodel_comm_history_add_entry_(
                                                   const uint64_t key,
                                                   const int src_dev, const int src_type,
                                                   const int dst_dev, const int dst_type,
                                                   const double mean,
                                                   const double sd,
                                                   const double sum,
                                                   const double sum_pow2,
                                                   const unsigned int samples
                                                   )
{
  kaapi_perfmodel_comm_history_t* perfmodel_comm;

  perfmodel_comm = kaapi_perfmodel_comm_history_find_entry(&kaapi_perfmodel_comm_hash[src_type][dst_type], key);
  perfmodel_comm->dev[src_dev][dst_dev].mean = mean;
  perfmodel_comm->dev[src_dev][dst_dev].sd = sd;
  perfmodel_comm->dev[src_dev][dst_dev].sum = sum;
  perfmodel_comm->dev[src_dev][dst_dev].sum_pow2 = sum_pow2;
  perfmodel_comm->dev[src_dev][dst_dev].samples = samples;
}

void kaapi_perfmodel_comm_history_print_stats(void)
{
#if defined(KAAPI_DEBUG)
  static const uint32_t map_size = KAAPI_HASHMAP_BIG_SIZE;
  kaapi_hashentries_t *entry;
  uint32_t i;
  int src, dst;
  int src_type, dst_type;

  for(src_type = 0; src_type < KAAPI_PROC_TYPE_MAX; src_type++) {
    for(dst_type = 0; dst_type < KAAPI_PROC_TYPE_MAX; dst_type++){
      kaapi_big_hashmap_t *const hmap = &kaapi_perfmodel_comm_hash[src_type][dst_type].hash;
      
      for (i = 0; i < map_size; i++) {
        for (entry = hmap->entries[i]; entry != NULL; entry = entry->next) {
          kaapi_perfmodel_comm_history_t* perfmodel_comm = (kaapi_perfmodel_comm_history_t*)entry->u.perf;
          if (perfmodel_comm == 0)
            continue;
          
          if(perfmodel_comm->key == 0)
            continue;
          
          for(src= 0; src < KAAPI_MAX_PROCESSOR_LIMIT; src++){
            for(dst= 0; dst < KAAPI_MAX_PROCESSOR_LIMIT; dst++){
              if(perfmodel_comm->dev[src][dst].samples == 0)
                continue;
              
              fprintf(stdout, "[%s] (%d:%s -> %d:%s) key=%x samples=%d mean=%.10f sd=%.10f\n",
                      __FUNCTION__,
                      src, (src_type == KAAPI_MEM_TYPE_CUDA)?"CUDA":"CPU",
                      dst, (dst_type == KAAPI_MEM_TYPE_CUDA)?"CUDA":"CPU",
                      perfmodel_comm->key,
                      perfmodel_comm->dev[src][dst].samples,
                      perfmodel_comm->dev[src][dst].mean,
                      perfmodel_comm->dev[src][dst].sd
                      );
              fflush(stdout);
            }
          }
        }
      }
    }
  }
#endif
}

static inline void kaapi_perfmodel_comm_history_file_dump(FILE* file)
{
  static const uint32_t map_size = KAAPI_HASHMAP_BIG_SIZE;
  kaapi_hashentries_t *entry;
  uint32_t i;
  int src, dst;
  int src_type, dst_type;
  
  for(src_type = 0; src_type < KAAPI_PROC_TYPE_MAX; src_type++) {
    for(dst_type = 0; dst_type < KAAPI_PROC_TYPE_MAX; dst_type++){
      kaapi_big_hashmap_t *const hmap = &kaapi_perfmodel_comm_hash[src_type][dst_type].hash;
      
      for (i = 0; i < map_size; i++) {
        for (entry = hmap->entries[i]; entry != NULL; entry = entry->next) {
          kaapi_perfmodel_comm_history_t* perfmodel_comm = (kaapi_perfmodel_comm_history_t*)entry->u.perf;
          if (perfmodel_comm == 0)
            continue;
          
          if(perfmodel_comm->key == 0)
            continue;
          
          for(src= 0; src < KAAPI_MAX_PROCESSOR_LIMIT; src++){
            for(dst= 0; dst < KAAPI_MAX_PROCESSOR_LIMIT; dst++){
              if(perfmodel_comm->dev[src][dst].samples == 0)
                continue;
              // ( src(id, type) | dest(id, type) | size )
              fprintf(file, "%x %d %d %d %d %le %le %le %le %d\n",
                      perfmodel_comm->key,
                      src, src_type,
                      dst, dst_type,
                      perfmodel_comm->dev[src][dst].mean,
                      perfmodel_comm->dev[src][dst].sd,
                      perfmodel_comm->dev[src][dst].sum,
                      perfmodel_comm->dev[src][dst].sum_pow2,
                      perfmodel_comm->dev[src][dst].samples
                      );
              fflush(stdout);
            }
          }
        }
      }
    }
  }  
}

static inline FILE* kaapi_perfmodel_comm_history_file_open_w(void)
{
  FILE* file = fopen(KAAPI_PERFMODEL_COMM_FILE, "w+");
  kaapi_assert(file != NULL);
  return file;
}

static inline void kaapi_perfmodel_comm_history_file_close(FILE* file)
{
  fclose(file);
}

static inline FILE* kaapi_perfmodel_comm_history_file_open_r(void)
{
  FILE* file = fopen(KAAPI_PERFMODEL_COMM_FILE, "r");
  return file;
}

static inline void kaapi_perfmodel_comm_history_file_parse(FILE* file)
{
  uint32_t key;
  int src, src_type, dst, dst_type;
  uint64_t samples;
  double mean, sd, sum, sum_pow2;
  int ret;
  
  while(1)
  {
    ret = fscanf(file, "%x %d %d %d %d %le %le %le %le %d\n",
                 &key,
                 &src, &src_type,
                 &dst, &dst_type,
                 &mean,
                 &sd,
                 &sum,
                 &sum_pow2,
                 &samples
                 );
    if((ret == EOF) || (ret != KAAPI_PERFMODEL_MAX_PARAMS))
      break;
    
    kaapi_perfmodel_comm_history_add_entry_(key, src, src_type, dst, dst_type, mean, sd, sum, sum_pow2, samples);
  }
}

/**************************************************************************
 null Communication model - always returns zero.
 */
static inline void kaapi_perfmodel_comm_null_init(void)
{
}

static inline void kaapi_perfmodel_comm_null_finalize(void)
{
}

static inline double kaapi_perfmodel_comm_null_predict(kaapi_processor_t* const kproc, kaapi_taskdescr_t* const td)
{
  return 0.f;
}

static inline void kaapi_perfmodel_comm_null_update(const kaapi_address_space_id_t src_asid, const kaapi_address_space_id_t dst_asid,
                                      const size_t size, const double time)
{
  
}

static inline void kaapi_perfmodel_history_hash_init(kaapi_perfmodel_comm_history_hash_t* hash)
{
  kaapi_big_hashmap_init(&hash->hash, 0);
  kaapi_atomic_initlock(&hash->lock);
}

static inline void kaapi_perfmodel_history_hash_destroy(kaapi_perfmodel_comm_history_hash_t* hash)
{
  kaapi_big_hashmap_destroy(&hash->hash);
  kaapi_atomic_destroylock(&hash->lock);
}

/**************************************************************************
 History Communication model - reads from a file the predicted performance by entry (src,dst,size).
 Entry: ( src(id, type) | dest(id, type) | size )
 */
static inline void kaapi_perfmodel_comm_history_init(void)
{
  int i, j;
  FILE* file;
  
  for(i = 0; i < KAAPI_PROC_TYPE_MAX; i++) {
    for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++){
      kaapi_perfmodel_history_hash_init(&kaapi_perfmodel_comm_hash[i][j]);
    }
  }
  
  file = kaapi_perfmodel_comm_history_file_open_r();
  if(file != 0)
  {
    kaapi_perfmodel_comm_history_file_parse(file);
    kaapi_perfmodel_comm_history_file_close(file);
//    kaapi_perfmodel_comm_history_print_stats();
  }
}

static inline void kaapi_perfmodel_comm_history_finalize(void)
{
  int i, j;

//  kaapi_perfmodel_comm_history_print_stats();
  if(kaapi_perfmodel_calibrate_is_enabled())
  {
    FILE* file = kaapi_perfmodel_comm_history_file_open_w();
    if(file != 0)
    {
      kaapi_perfmodel_comm_history_file_dump(file);
      kaapi_perfmodel_comm_history_file_close(file);
    }
  }

  for(i = 0; i < KAAPI_PROC_TYPE_MAX; i++) {
    for(j= 0; j < KAAPI_PROC_TYPE_MAX; j++){
      kaapi_perfmodel_history_hash_destroy(&kaapi_perfmodel_comm_hash[i][j]);
    }
  }
}

static inline void kaapi_perfmodel_comm_history_add_sample(kaapi_perfmodel_comm_history_t* const entry,
                                                           const int src, const int dst,
                                                           const double time)
{
  kaapi_atomic_lock(&entry->dev[src][dst].lock);
  entry->dev[src][dst].samples++;
  entry->dev[src][dst].sum += time;
  entry->dev[src][dst].sum_pow2 += (time*time);
  
  unsigned int n = entry->dev[src][dst].samples;
  entry->dev[src][dst].mean = entry->dev[src][dst].sum / n;
  entry->dev[src][dst].sd =
    sqrt( (entry->dev[src][dst].sum_pow2 - (entry->dev[src][dst].sum*entry->dev[src][dst].sum)/n)/n );
  kaapi_atomic_unlock(&entry->dev[src][dst].lock);
}

static inline int kaapi_perfmodel_comm_history_get_dev(const kaapi_address_space_id_t asid)
{
  if(kaapi_memory_address_space_gettype(asid) == KAAPI_MEM_TYPE_CUDA)
  {
#if defined(KAAPI_USE_CUDA)
    kaapi_processor_t* const kproc = kaapi_all_kprocessors[kaapi_memory_map_asid2kid(asid)];
    return kaapi_processor_get_cudaproc(kproc)->index;
#else
    return 0;
#endif
  } else {
    kaapi_processor_t* const kproc = kaapi_get_current_processor();
    return kaapi_default_param.kid2cpu[kproc->kid];
  }
}

static inline void kaapi_perfmodel_comm_history_update_entry(const kaapi_address_space_id_t src_asid, const kaapi_address_space_id_t dst_asid,
                                               const size_t size, const double time)
{
  if(kaapi_perfmodel_calibrate_is_enabled())
  {
    const int src_arch = kaapi_memory_address_space_gettype(src_asid);
    const int dst_arch = kaapi_memory_address_space_gettype(dst_asid);
    const int src_dev = kaapi_perfmodel_comm_history_get_dev(src_asid);
    const int dst_dev = kaapi_perfmodel_comm_history_get_dev(dst_asid);
    const uint64_t key = (uint64_t)size;
//    fprintf(stdout, "%d:%s: src=%d(%s) dst=%d(%s) size=%d time=%.4f\n", kaapi_get_self_kid(), __FUNCTION__,
//            src_dev, (src_arch == KAAPI_MEM_TYPE_CUDA)?"CUDA":"CPU",
//            dst_dev, (dst_arch == KAAPI_MEM_TYPE_CUDA)?"CUDA":"CPU",
//            size, time);
//    fflush(stdout);
    kaapi_perfmodel_comm_history_hash_t* hash = &kaapi_perfmodel_comm_hash[src_arch][dst_arch];
    kaapi_perfmodel_comm_history_t* perfmodel_comm;
    perfmodel_comm = kaapi_perfmodel_comm_history_find_entry(hash, key);
    kaapi_perfmodel_comm_history_add_sample(perfmodel_comm, src_dev, dst_dev, time);
  }
}

static inline int kaapi_perfmodel_comm_history_get_dev_(kaapi_processor_t* const kproc)
{
  if(kaapi_processor_get_type(kproc) == KAAPI_PROC_TYPE_CUDA)
  {
#if defined(KAAPI_USE_CUDA)
    return kaapi_processor_get_cudaproc(kproc)->index;
#else
    return 0;
#endif
  } else {
    return kaapi_default_param.kid2cpu[kproc->kid];
  }
}

static inline double kaapi_perfmodel_comm_history_get_predict(kaapi_processor_t* const kdest, kaapi_processor_t* const ksrc, uint64_t nbytes)
{
  uint64_t key = (uint64_t)nbytes;
  const int dst_type = kaapi_processor_get_type(kdest);
  const int src_type = kaapi_processor_get_type(ksrc);
  
  const int dst_dev = kaapi_perfmodel_comm_history_get_dev_(kdest);
  const int src_dev = kaapi_perfmodel_comm_history_get_dev_(ksrc);

  kaapi_assert_debug( (src_type < KAAPI_PROC_TYPE_MAX) && (dst_type < KAAPI_PROC_TYPE_MAX) );
  kaapi_assert_debug( (src_dev < KAAPI_MAX_PROCESSOR_LIMIT) && (dst_dev < KAAPI_MAX_PROCESSOR_LIMIT) );
  
  kaapi_perfmodel_comm_history_hash_t* const hash = &kaapi_perfmodel_comm_hash[src_type][dst_type];
  kaapi_perfmodel_comm_history_t* const perfmodel_comm = kaapi_perfmodel_comm_history_find_entry(hash, key);
  
  return perfmodel_comm->dev[src_dev][dst_dev].mean;
}

static inline double kaapi_perfmodel_comm_history_predict(kaapi_processor_t* const kproc, kaapi_taskdescr_t* const td)
{
  size_t i;
  void* sp = td->task->sp;
  const size_t count_params = kaapi_format_get_count_params(td->fmt, sp);
  kaapi_address_space_id_t kasid = kaapi_memory_map_kid2asid(kproc->kid);
  uint64_t nbytes = 0;
  double predicted= 0.f, timeus;
  
  for (i = 0; i < count_params; i++) {
    kaapi_access_mode_t m = kaapi_format_get_mode_param(td->fmt, i, sp);
    
    m = KAAPI_ACCESS_GET_MODE(m);
    if (m == KAAPI_ACCESS_MODE_V)
      continue;
    
    kaapi_access_t access = kaapi_format_get_access_param(td->fmt, i, sp);
    kaapi_data_t* const kdata = kaapi_data(kaapi_data_t, &access);
    kaapi_metadata_info_t* const kmdi = kdata->mdi;
    kaapi_assert_debug(kmdi != 0);
    
    if(!kaapi_metadata_info_is_valid(kmdi, kasid))
    {
      nbytes = kaapi_memory_view_size(&kdata->view);
      uint16_t lid = kaapi_metadata_info_first_valid(kmdi);
      kaapi_processor_t* const ksrc =  kaapi_all_kprocessors[kaapi_memory_map_lid2kid(lid)];
      timeus = kaapi_perfmodel_comm_history_get_predict(kproc, ksrc, nbytes);
      if(timeus > 0.f) {
        predicted += timeus;
      }
    }
  }
  
  return predicted;
}

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_idgraf = {
  .init                = kaapi_perfmodel_comm_idgraf_init,
  .finalize            = kaapi_perfmodel_comm_idgraf_finalize,
  .predict             = kaapi_perfmodel_comm_idgraf_predict,
  .update              = kaapi_perfmodel_comm_null_update
};

/*
 Format:
 
 */
static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_history = {
  .init                = kaapi_perfmodel_comm_history_init,
  .finalize            = kaapi_perfmodel_comm_history_finalize,
  .predict             = kaapi_perfmodel_comm_history_predict,
  .update              = kaapi_perfmodel_comm_history_update_entry
};

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_null = {
  .init                = kaapi_perfmodel_comm_null_init,
  .finalize            = kaapi_perfmodel_comm_null_finalize,
  .predict             = kaapi_perfmodel_comm_null_predict,
  .update              = kaapi_perfmodel_comm_null_update
};

void kaapi_perfmodel_comm_init(void)
{
  const char* perfmodel = getenv("KAAPI_PERFMODEL_TRANSFER");
  
  kaapi_perfmodel_comm = kaapi_perfmodel_comm_null;
  if( perfmodel != 0 )
  {
    if (strcmp(perfmodel, "idgraf") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_idgraf;
    else if (strcmp(perfmodel, "history") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_history;
    else if (strcmp(perfmodel, "null") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_null;
    else {
      fprintf(stderr, "*** Kaapi: bad value for 'KAAPI_PERFMODEL_TRANSFER': '%s'\n",
              getenv("KAAPI_PERFMODEL_TRANSFER")
              );
    }
  }
  
  kaapi_perfmodel_comm.init();
}

void kaapi_perfmodel_comm_finalize(void)
{
  kaapi_perfmodel_comm.finalize();
}

double kaapi_perfmodel_comm_predict_transfer(kaapi_processor_t* const kproc, kaapi_taskdescr_t* const td)
{
  return kaapi_perfmodel_comm.predict(kproc, td);
}

void kaapi_perfmodel_comm_update(const kaapi_address_space_id_t src_asid, const kaapi_address_space_id_t dst_asid,
                                        const size_t size, const double time)
{
  kaapi_perfmodel_comm.update(src_asid, dst_asid, size, time);
}

==> xkaapi-2.2.tar.gz: xkaapi-2.2/src/perfmodel/kaapi_perfmodel_comm.c <==

#include <stdio.h>
#include <math.h>

#include "kaapi_impl.h"
#include "kaapi_tasklist.h"

#include "kaapi_perfmodel_idgraf.h"

#if defined(KAAPI_USE_CUDA)
#include "machine/cuda/kaapi_cuda_impl.h"
#endif

#include "kaapi_perfmodel_bandwidth.h"
#include "kaapi_perfmodel_history.h"

/** return time in microseconds (us)
 */
static inline double kaapi_get_elapsedus(void)
{
  uint64_t retval;
  retval = kaapi_get_elapsedns();
  return (double)(retval / 1e3);
}

typedef struct kaapi_perfmodel_comm_t {
  void      (*init)       (void);
  void      (*finalize)   (void);
  double    (*predict)    (kaapi_processor_t* const, kaapi_taskdescr_t* const);
  void      (*update)     (const kaapi_address_space_id_t, const kaapi_address_space_id_t, const size_t, const double);
} kaapi_perfmodel_comm_t;

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm;

/**************************************************************************
 null Communication model - always returns zero.
 */
static inline void kaapi_perfmodel_comm_null_init(void)
{
}

static inline void kaapi_perfmodel_comm_null_finalize(void)
{
}

static inline double kaapi_perfmodel_comm_null_predict(kaapi_processor_t* const kproc, kaapi_taskdescr_t* const td)
{
  return 0.f;
}

static inline void kaapi_perfmodel_comm_null_update(const kaapi_address_space_id_t src_asid, const kaapi_address_space_id_t dst_asid,
                                      const size_t size, const double time)
{
  
}

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_idgraf = {
  .init                = kaapi_perfmodel_comm_idgraf_init,
  .finalize            = kaapi_perfmodel_comm_idgraf_finalize,
  .predict             = kaapi_perfmodel_comm_idgraf_predict,
  .update              = kaapi_perfmodel_comm_null_update
};

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_history = {
  .init                = kaapi_perfmodel_comm_history_init,
  .finalize            = kaapi_perfmodel_comm_history_finalize,
  .predict             = kaapi_perfmodel_comm_history_predict,
  .update              = kaapi_perfmodel_comm_history_update_entry
};

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_bandwidth = {
  .init                = kaapi_perfmodel_comm_bandwidth_init,
  .finalize            = kaapi_perfmodel_comm_bandwidth_finalize,
  .predict             = kaapi_perfmodel_comm_bandwidth_predict,
  .update              = 0
};

static kaapi_perfmodel_comm_t kaapi_perfmodel_comm_null = {
  .init                = kaapi_perfmodel_comm_null_init,
  .finalize            = kaapi_perfmodel_comm_null_finalize,
  .predict             = kaapi_perfmodel_comm_null_predict,
  .update              = kaapi_perfmodel_comm_null_update
};

void kaapi_perfmodel_comm_init(void)
{
  const char* perfmodel = getenv("KAAPI_PERFMODEL_TRANSFER");
  
  kaapi_perfmodel_comm = kaapi_perfmodel_comm_null;
  if( perfmodel != 0 )
  {
    if (strcmp(perfmodel, "idgraf") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_idgraf;
    else if (strcmp(perfmodel, "history") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_history;
    else if (strcmp(perfmodel, "bandwidth") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_bandwidth;
    else if (strcmp(perfmodel, "null") == 0)
      kaapi_perfmodel_comm = kaapi_perfmodel_comm_null;
    else {
      fprintf(stderr, "*** Kaapi: bad value for 'KAAPI_PERFMODEL_TRANSFER': '%s'\n",
              getenv("KAAPI_PERFMODEL_TRANSFER")
              );
    }
  }
  
  if(kaapi_perfmodel_comm.init != 0)
    kaapi_perfmodel_comm.init();
}

void kaapi_perfmodel_comm_finalize(void)
{
  if(kaapi_perfmodel_comm.finalize != 0)
    kaapi_perfmodel_comm.finalize();
}

double kaapi_perfmodel_comm_predict_transfer(kaapi_processor_t* const kproc, kaapi_taskdescr_t* const td)
{
  if(kaapi_perfmodel_comm.predict != 0)
    return kaapi_perfmodel_comm.predict(kproc, td);
  
  return 0.f;
}

void kaapi_perfmodel_comm_update(const kaapi_address_space_id_t src_asid, const kaapi_address_space_id_t dst_asid,
                                        const size_t size, const double time)
{
  if(kaapi_perfmodel_comm.update != 0)
    kaapi_perfmodel_comm.update(src_asid, dst_asid, size, time);
}
